(*
 ________                        _____________
 __  ___/______ ____   _______   ___  __/___(_)_______ ___ _____
 _____ \ _  __ `/__ | / /_  _ \  __  /   __  / __  __ `__ \_  _ \
 ____/ / / /_/ / __ |/ / /  __/  _  /    _  /  _  / / / / //  __/
 /____/  \__,_/  _____/  \___/   /_/     /_/   /_/ /_/ /_/ \___/
 ___    _________________    ________              ______
 __ |  / /__  ____/___  /    ___  __/______ ______ ___  /________
 __ | / / _  /     __  /     __  /   _  __ \_  __ \__  / __  ___/
 __ |/ /  / /___   _  /___   _  /    / /_/ // /_/ /_  /  _(__  )
 _____/   \____/   /_____/   /_/     \____/ \____/ /_/   /____/

 Delphi components with source codes (c) Jakub Dusek
 ----------------------------------------------------------------
 TSTColorsCombo & TSTColorsList 1.12
 ----------------------------------------------------------------
 Displays colors combo or list box with preview rectangle
 See included help files for detail description.
 ----------------------------------------------------------------
 This source code is not public domain and can be distributed
 only with this header. Mail me some bugs or good ideas about
 new functions.
 ----------------------------------------------------------------
 http://dusek.zde.cz                            j.dusek@seznam.cz
 ________________________________________________________________

*)

unit STColors;

{$IFNDEF VER80}
 {$IFNDEF VER90}
  {$IFNDEF VER93}
    {$IFNDEF VER100}
      {$IFNDEF VER110}
        {$DEFINE ST4}
        {$IFNDEF VER120}
           {$IFNDEF VER125}
            {$DEFINE ST5}
          {$ENDIF}
        {$ENDIF}
      {$ENDIF}
    {$ENDIF}
  {$ENDIF}
 {$ENDIF}
{$ENDIF}

interface

uses Windows, Graphics, StdCtrls, Classes, Messages,
     Controls, SysUtils;

const
  StdColors: array [0..15] of TColor =
  (clBlack, clMaroon, clGreen, clOlive,
  clNavy, clPurple, clTeal, clGray,
  clSilver, clRed, clLime, clYellow,
  clBlue, clFuchsia, clAqua, clWhite);

  SysColors: array [0..24] of TColor =
  (clScrollBar, clBackground, clActiveCaption,
  clInactiveCaption, clMenu, clWindow, clWindowFrame,
  clMenuText, clWindowText, clCaptionText, clActiveBorder,
  clInactiveBorder, clAppWorkSpace, clHighlight,
  clHighlightText, clBtnFace, clBtnShadow, clGrayText,
  clBtnText, clInactiveCaptionText, clBtnHighlight,
  cl3DDkShadow, cl3DLight, clInfoText, clInfoBk);

type
  TSTColorsCombo = class (TCustomComboBox)
  private
    FColorNames		: TStringList;
    FColorHeight	: integer;
    FColorSelected	: TColor;
    FColorText		: string;
    FColorWidth		: integer;
    FDisplayNames	: Boolean;
    FSystemColors	: Boolean;
    procedure SetColorNames (Value: TStringList);
    procedure SetColorHeight (Value: integer);
    procedure SetColorSelected (Value: TColor);
    procedure SetColorText (Value: string);
    procedure SetColorWidth (Value: integer);
    procedure SetDisplayNames (Value: Boolean);
    procedure SetSystemColors (Value: Boolean);
    procedure AddColors;
    procedure NamesChanged (Sender: TObject);
    procedure CMFontChanged (var Message: TMessage); message CM_FONTCHANGED;
    procedure WMSize (var Message: TMessage); message WM_SIZE;
  protected
    procedure Change; override;
    procedure DrawItem (Index: Integer; Rect: TRect; State: TOwnerDrawState); override;
  public
    constructor Create (AOwner:TComponent); override;
    procedure CreateWnd; override;
    procedure Loaded; override;
    destructor Destroy; override;
  published
    property ColorNames		: TStringList read FColorNames write SetColorNames;
    property ColorHeight	: integer read FColorHeight write SetColorHeight;
    property ColorSelected	: TColor read FColorSelected write SetColorSelected;
    property ColorText		: string read FColorText write SetColorText;
    property ColorWidth		: integer read FColorWidth write SetColorWidth default 22;
    property DisplayNames	: Boolean read FDisplayNames write SetDisplayNames default True;
    property SystemColors	: Boolean read FSystemColors write SetSystemColors default True;
    property Items;
   {$IFDEF ST4}
    property Anchors;
    property BiDiMode;
    property Constraints;
    property DragKind;
    property ParentBiDiMode;
    property OnEndDock;
    property OnStartDock;
    {$ENDIF}
    {$IFDEF ST5}
    property OnContextPopup;
    {$ENDIF}
    {$IFNDEF VER90}
    property ImeMode;
    property ImeName;
    {$ENDIF}
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property DropDownCount;
    property Enabled;
    property Font;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDropDown;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnStartDrag;
  end;

  TSTColorsList = class (TCustomListBox)
  private
    FColorNames		: TStringList;
    FColorHeight	: integer;
    FColorSelected	: TColor;
    FColorText		: string;
    FColorWidth		: integer;
    FDisplayNames	: Boolean;
    FSystemColors	: Boolean;
    procedure SetColorNames (Value: TStringList);
    procedure SetColorHeight (Value: integer);
    procedure SetColorSelected (Value: TColor);
    procedure SetColorText (Value: string);
    procedure SetColorWidth (Value: integer);
    procedure SetDisplayNames (Value: Boolean);
    procedure SetSystemColors (Value: Boolean);
    procedure AddColors;
    procedure NamesChanged (Sender: TObject);
    procedure CMFontChanged (var Message: TMessage); message CM_FONTCHANGED;
    procedure WMSize (var Message: TMessage); message WM_SIZE;
  protected
    procedure Click; override;
    procedure DrawItem (Index: Integer; Rect: TRect; State: TOwnerDrawState); override;
  public
    constructor Create (AOwner:TComponent); override;
    procedure CreateWnd; override;
    procedure Loaded; override;
    destructor Destroy; override;
  published
    property ColorNames		: TStringList read FColorNames write SetColorNames;
    property ColorHeight	: integer read FColorHeight write SetColorHeight;
    property ColorSelected	: TColor read FColorSelected write SetColorSelected;
    property ColorText		: string read FColorText write SetColorText;
    property ColorWidth		: integer read FColorWidth write SetColorWidth default 22;
    property DisplayNames	: Boolean read FDisplayNames write SetDisplayNames default True;
    property SystemColors	: Boolean read FSystemColors write SetSystemColors default True;
    {$IFDEF ST4}
    property Anchors;
    property BiDiMode;
    property Constraints;
    property DragKind;
    property ParentBiDiMode;
    property OnEndDock;
    property OnStartDock;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    {$ENDIF}
    {$IFDEF ST5}
    property OnConTextPopup;
    {$ENDIF}
    {$IFNDEF VER90}
    property ImeMode;
    property ImeName;
    {$ENDIF}
    property Align;
    property BorderStyle;
    property Color;
    property Columns;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property ExtendedSelect;
    property Font;
    property IntegralHeight;
    property MultiSelect;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnStartDrag;
  end;

  procedure Register;

implementation

procedure Register;
begin
  RegisterComponents ('ST Tools', [TSTColorsCombo]);
  RegisterComponents ('ST Tools', [TSTColorsList]);
end;

function GetContrastColor (Color: TColor): TColor;
var r, g, b: Byte;
begin
  Color := ColorToRGB (Color);
  if GetRValue (Color) >= $80 then r := 0
  else
    r := $FF;
  if GetGValue (Color) >= $80 then g := 0
  else
    g := $FF;
  if GetRValue (Color) >= $80 then b := 0
  else
    b := $FF;
  Result := RGB (r,g,b);
end;

function GetItemHeight (Font: TFont): Integer;
var
  DC		: HDC;
  Fnt		: HFont;
  Metrics	: TTextMetric;
begin
  DC := GetDC (0);
  try
    Fnt := SelectObject (DC, Font.Handle);
    GetTextMetrics (DC, Metrics);
    SelectObject (DC, Fnt);
  finally
    ReleaseDC (0, DC);
  end;
  Result := Metrics.tmHeight + 1;
end;

procedure DrawColorItem (Index: integer; Canvas: TCanvas; Rect: TRect; Font: TFont;
                         DisplayNames: Boolean; ColorWidth: Integer;
                         Text, ColorText: string; Color: TColor; ColorNames: TStringList);
var r		: TRect;
    s, t 	: string;
    c, fc	: TColor;
begin
  r := Rect;
  Canvas.FillRect (Rect);
  Inc (r.Top, 2);
  Inc (r.Left, 2);
  Dec (r.Bottom, 2);
  if not DisplayNames then Dec (r.Right, 2)
  else
    r.Right := r.Left + ColorWidth;
  c := Canvas.Brush.Color;
  Canvas.Brush.Color := Color;
  Canvas.Rectangle (r.Left, r.Top, r.Right, r.Bottom);
  s := Text;
  if ColorNames.Count < Index + 1 then
  begin
    if LowerCase (Copy (s, 1, 2)) = 'cl' then s := Copy (s, 3, Length (s));
  end
  else
  begin
    s := ColorNames [index];
  end;
  if ColorText <> '' then
  begin
    fc := Canvas.Font.Color;
    Canvas.Font.Color := GetContrastColor (Canvas.Brush.Color);
    t := ColorText;
    if t = '%1' then t := s;
    DrawText (Canvas.Handle, PChar(t), Length(t), R,
              DT_SINGLELINE or DT_VCENTER or DT_NOPREFIX or DT_CENTER);
    Canvas.Font.Color := fc;
  end;
  Canvas.Brush.Color := c;
  rect.Left := r.Right + 2;
  DrawText (Canvas.Handle, PChar(s), Length(s), Rect,
            DT_SINGLELINE or DT_VCENTER or DT_NOPREFIX);
end;

constructor TSTColorsCombo.Create(AOwner:TComponent);
begin
  inherited Create (AOwner);
  Style := csOwnerDrawFixed;
  FColorWidth := 22;
  FColorHeight := ItemHeight;
  FDisplayNames := True;
  FSystemColors := True;
  FColorNames := TStringList.Create;
  TStringList (FColorNames).OnChange := NamesChanged;
end;

procedure TSTColorsCombo.CreateWnd;
var i: integer;
begin
  i := GetItemHeight (Font);
  if i > FColorHeight then FColorHeight := i;
  Height := FColorHeight;
  ItemHeight := FColorHeight;
  inherited CreateWnd;
  Loaded;
end;

procedure TSTColorsCombo.Loaded;
begin
  inherited Loaded;
  FColorHeight := ItemHeight;
  SetColorSelected (FColorSelected);
  AddColors;
end;

destructor TSTColorsCombo.Destroy;
begin
  FColorNames.Free;
  inherited Destroy;
end;

procedure TSTColorsCombo.AddColors;
var i, Index	: Integer;
    s		: string;
begin
  if Items.Count <> 0 then s := Items [ItemIndex];
  Items.BeginUpdate;
  try
    Clear;
    for i := 0 to High (StdColors) do
    begin
      Items.AddObject (ColorToString (StdColors [i]), TObject (StdColors [i]));
    end;
    if FSystemColors then
    begin
      for i := 0 to High (SysColors) do
      begin
        Items.AddObject (ColorToString (SysColors [i]), TObject (SysColors [i]));
      end;
    end;
  finally
    Items.EndUpdate;
  end;

  Index := Items.IndexOf (s);
  if Index <> -1 then ItemIndex := Index
  else
    ItemIndex := 0;
  SetColorSelected (StringToColor (Items [ItemIndex]));

end;

procedure TSTColorsCombo.SetColorNames (Value: TStringList);
begin
  FColorNames.Assign (Value);
  Invalidate;
end;

procedure TSTColorsCombo.NamesChanged (Sender: TObject);
begin
  Invalidate;
end;

procedure TSTColorsCombo.SetColorWidth (Value: integer);
begin
  FColorWidth := Value;
  Invalidate;
end;

procedure TSTColorsCombo.SetSystemColors (Value: Boolean);
begin
  FSystemColors := Value;
  AddColors;
end;

procedure TSTColorsCombo.SetColorHeight (Value: integer);
begin
  FColorHeight := Value;
  RecreateWnd;
end;

procedure TSTColorsCombo.SetDisplayNames (Value: Boolean);
begin
  FDisplayNames := Value;
  Invalidate;
end;

procedure TSTColorsCombo.SetColorSelected (Value: TColor);
var Index: Integer;
begin
  Index := Items.IndexOf (ColorToString (Value));
  if Index <> -1 then
  begin
    FColorSelected := Value;
    ItemIndex := Index;
    if not (csReading in ComponentState) and
       Assigned (OnChange) then OnChange (Self);
  end;
end;

procedure TSTColorsCombo.SetColorText (Value: string);
begin
  FColorText := Value;
  Invalidate;
end;

procedure TSTColorsCombo.CMFontChanged (var Message: TMessage);
begin
  inherited;
  RecreateWnd;
end;

procedure TSTColorsCombo.WMSize (var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TSTColorsCombo.Change;
begin
  SetColorSelected (StringToColor (Items [ItemIndex]));
  inherited Change;
end;

procedure TSTColorsCombo.DrawItem (Index: Integer; Rect: TRect; State: TOwnerDrawState);
begin
  DrawColorItem (Index, Canvas,Rect, Font, FDisplayNames,
                 FColorWidth, Items[Index], FColorText,
                 TColor (Items.Objects [Index]), ColorNames);
end;

constructor TSTColorsList.Create (AOwner: TComponent);
begin
  inherited Create (AOwner);
  Style := lbOwnerDrawFixed;
  FColorWidth := 22;
  FColorHeight := ItemHeight;
  FDisplayNames := True;
  FSystemColors := True;
  FColorNames := TStringList.Create;
  TStringList (FColorNames).OnChange := NamesChanged;
end;

procedure TSTColorsList.CreateWnd;
var i: integer;
begin
  i := GetItemHeight (Font);
  if i > FColorHeight then FColorHeight := i;
  ItemHeight := FColorHeight;
  inherited CreateWnd;
  AddColors;
end;

procedure TSTColorsList.Loaded;
begin
  inherited Loaded;
  FColorHeight := ItemHeight;
  SetColorSelected (FColorSelected);
end;

destructor TSTColorsList.Destroy;
begin
  FColorNames.Free;
  inherited Destroy;
end;

procedure TSTColorsList.AddColors;
var i, Index	: integer;
    s		: string;
begin
  if Items.Count <> 0 then s := Items [ItemIndex];
  Items.Clear;
  for i := 0 to High (StdColors) do
  begin
    Items.AddObject (ColorToString (StdColors [i]), TObject (StdColors [i]));
  end;
  if FSystemColors then
  begin
    for i := 0 to High (SysColors) do
    begin
      Items.AddObject (ColorToString (SysColors [i]), TObject (SysColors [i]));
    end;
  end;
  Index := Items.indexof(s);
  if Index <> -1 then ItemIndex := Index
  else
    ItemIndex := 0;
  SetColorSelected (StringToColor (Items [ItemIndex]));
end;

procedure TSTColorsList.SetColorNames (Value: TStringList);
begin
  FColorNames.Assign (Value);
  Invalidate;
end;

procedure TSTColorsList.NamesChanged (Sender: TObject);
begin
  Invalidate;
end;

procedure TSTColorsList.SetColorWidth (Value: integer);
begin
  FColorWidth := Value;
  Invalidate;
end;

procedure TSTColorsList.SetSystemColors (Value: Boolean);
begin
  FSystemColors := Value;
  AddColors;
end;

procedure TSTColorsList.SetColorHeight (Value: integer);
begin
  FColorHeight := Value;
  RecreateWnd;
end;

procedure TSTColorsList.SetDisplayNames (Value: Boolean);
begin
  FDisplayNames := Value;
  Invalidate;
end;

procedure TSTColorsList.SetColorSelected (Value: TColor);
var Index: Integer;
begin
  Index := Items.IndexOf (ColorToString (Value)); 
  if Index <> -1 then
  begin
    FColorSelected := Value;
    ItemIndex := Index;
    if not (csReading in ComponentState) and
       Assigned (OnClick) then OnClick (Self);
  end;
end;

procedure TSTColorsList.SetColorText (Value: string);
begin
  FColorText := Value;
  Invalidate;
end;

procedure TSTColorsList.CMFontChanged (var Message: TMessage);
begin
  inherited;
  RecreateWnd;
end;

procedure TSTColorsList.WMSize (var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TSTColorsList.Click;
begin
  SetColorSelected (StringToColor (Items [ItemIndex]));
  inherited Click;
end;

procedure TSTColorsList.DrawItem (Index: Integer; Rect: TRect; State: TOwnerDrawState);
begin
  DrawColorItem (Index, Canvas, Rect, Font, FDisplayNames,
                 FColorWidth, Items[Index], FColorText,
                 TColor (Items.Objects [Index]), ColorNames);
end;

end.